from __future__ import annotations

import json
from dataclasses import dataclass, asdict
from typing import Any, AsyncIterator

import httpx

from . import Provider, ProviderError, Setting, system_prompt


def dataclass_to_json(obj: Any) -> dict[str, Any]:
    """Convert dataclass to a json dict

    This function filters out 'None' values.

    :param obj: the dataclass to serialize
    :return: serialized dataclass
    :raises TypeError: if obj is not a dataclass
    """
    return {k: v for k, v in asdict(obj).items() if v is not None}


@dataclass
class GenerateRequest:
    model: str
    """(required) the model name"""
    prompt: str | None = None
    """the prompt to generate a response for"""
    images: list[str] | None = None
    """a list of base64-encoded images (for multimodal models such as llava)"""
    format: str | None = None
    """the format to return a response in. Currently the only accepted value is json"""
    options: dict | None = None
    """additional model parameters listed in the documentation for the Modelfile such as temperature"""
    system: str | None = None
    """system prompt to (overrides what is defined in the Modelfile)"""
    template: str | None = None
    """the full prompt or prompt template (overrides what is defined in the Modelfile)"""
    context: str | None = None
    """the context parameter returned from a previous request to /generate, this can be used to keep a short 
    conversational memory"""
    stream: bool | None = None
    """if false the response will be returned as a single response object, rather than a stream of objects"""
    raw: bool | None = None
    """if true no formatting will be applied to the prompt and no context will be returned. You may choose to use 
    the raw parameter if you are specifying a full templated prompt in your request to the API, and are managing 
    history yourself. JSON mode"""


class Ollama(Provider):
    name = "Ollama"

    host = Setting(default="localhost")
    port = Setting(default=11434)
    model = Setting(default="codellama:13b")

    @property
    def endpoint(self) -> str:
        # computed property because python descriptors need to be bound to an instance before access
        return f"http://{self.host}:{self.port}/api/generate"

    async def generate(self, prompt: str) -> AsyncIterator[str]:
        request = GenerateRequest(self.model, prompt, system=system_prompt, stream=True)
        data = dataclass_to_json(request)
        try:
            async with httpx.AsyncClient() as client:
                async with client.stream("POST", self.endpoint, json=data, timeout=20.0) as stream:
                    async for line in stream.aiter_lines():
                        response = json.loads(line)
                        if "error" in response:
                            raise ProviderError(response["error"])
                        yield response["response"]
        except (httpx.HTTPError, httpx.StreamError) as e:
            raise ProviderError(f"Something went wrong while querying Ollama: {e}") from e



from __future__ import annotations

from abc import abstractmethod
from collections.abc import AsyncIterator
from typing import Protocol, TypeVar, Generic, runtime_checkable

from ..config import get_config

system_prompt = (
    "Based on the following user description, generate a corresponding Bash command. Focus solely "
    "on interpreting the requirements and translating them into a single, executable Bash command. "
    "Ensure accuracy and relevance to the user's description. The output should be a valid Bash "
    "command that directly aligns with the user's intent, ready for execution in a command-line "
    "environment. Output nothing except for the command. No code block, no English explanation, "
    "no start/end tags."
)


class ProviderError(Exception):
    """LLM providers raise this error to gracefully indicate something has gone wrong."""


@runtime_checkable
class Provider(Protocol):
    """
    LLM Provider Protocol

    All LLM backends must implement this interface.
    """
    name: str

    @abstractmethod
    def generate(self, prompt: str) -> AsyncIterator[str]:
        """
        This is an asynchronous generator method which defines the protocol that a provider implementation
        should adhere to. The method takes a prompt as an argument and produces an asynchronous stream
        of string results.

        :param prompt: A string value which serves as input to the provider's process of generating results.
        :return: An asynchronous generator yielding string results.
        """
        # If you are wondering why the 'generate' signature doesn't include 'async', see
        # https://mypy.readthedocs.io/en/stable/more_types.html#asynchronous-iterators


T = TypeVar("T")


class Setting(Generic[T]):
    def __init__(self, *, name: str | None = None, default: T | None = None) -> None:
        self.name = name
        self.default = default

    def __set_name__(self, owner: type[Provider], name: str) -> None:
        if not self.name:
            self.name = name

    def __get__(self, instance: Provider, owner: type[Provider]) -> T:
        if instance is None:
            # Accessing settings as a class attribute is not supported because it prevents
            # inspect.get_members from determining the object type
            raise AttributeError("Settings must be accessed through a provider instance.")
        config = get_config()
        return config["provider"][owner.name][self.name]


def _providers() -> dict[str, type[Provider]]:
    from .ollama import Ollama
    from .openai import OpenAI
    from .localai import LocalAI
    providers = {
        Ollama.name: Ollama,
        OpenAI.name: OpenAI,
        LocalAI.name: LocalAI
    }
    return providers


def get_provider(name: str) -> type[Provider]:
    """Imports and loads a requested provider

    :param name: the provider name
    :return: the requested provider
    """

    return _providers()[name]


def list_providers() -> list[str]:
    return list(_providers())
